use std::{path::PathBuf, sync::Arc};

use axum::{Router, AddExtensionLayer, http::{HeaderValue, header}};
use error::ApplicationError;
use hyper::{Request, Body};
use sibyl::{Environment, SessionPool};
use structopt::{StructOpt, lazy_static::lazy_static};

use config::Config;
use tokio::signal;
use tower_http::set_header::SetResponseHeaderLayer;

mod config;
mod error;
mod routes;
mod sql;
mod model;

#[derive(StructOpt)]
struct Opt {
    #[structopt(
        short,
        long,
        help = "config file to use",
        default_value = "./config.toml"
    )]
    config: PathBuf,
}

pub struct ServiceInner {
    pool: SessionPool<'static>
}

pub type Service = Arc<ServiceInner>;

lazy_static!{
    pub static ref ORACLE: Environment = sibyl::env().expect("Sibyl error");
}

#[tokio::main]
async fn main() -> Result<(), ApplicationError> {
    let opt = Opt::from_args();
    let config = std::fs::read(&opt.config)?;
    let config = toml::from_slice::<Config>(&config)?;

    let pool = ORACLE.create_session_pool(&config.db_name, &config.db_username, &config.db_password, 0, 1, 10).await?;
    let service: Service = Arc::new(ServiceInner{ pool });

    let app = Router::new()
        .nest("/api", routes::routes())
        .layer(AddExtensionLayer::new(service))
        .layer(SetResponseHeaderLayer::<_, Request<Body>>::if_not_present(
            header::ACCESS_CONTROL_ALLOW_ORIGIN,
            HeaderValue::from_static("*"),
        ));

    axum::Server::bind(&config.addr)
        .serve(app.into_make_service()).with_graceful_shutdown(shutdown()).await?;

    Ok(())
}

async fn shutdown() {
    match signal::ctrl_c().await {
        Ok(()) => {},
        Err(_) => {},
    }
}