use axum::{Json, extract::{Path, Extension, Query}, response::IntoResponse, Router, routing::BoxRoute, handler::{get, put}};
use hyper::StatusCode;

use crate::{Service, error::ServiceError, model::{BookingBody, Booking, BookingQuery}};

async fn bookings(
    Query(query): Query<BookingQuery>,
    Extension(service): Extension<Service>,
) -> Result<impl IntoResponse, ServiceError> {
    let bookings = match query.months {
        Some(months) => service.get_last_bookings(months).await?,
        None => service.get_bookings().await?,
    };
    Ok(Json(bookings))
}

async fn booking(
    Path(id): Path<i32>,
    Extension(service): Extension<Service>,
) -> Result<impl IntoResponse, ServiceError> {
    let booking = service.get_booking(id).await?
        .ok_or_else(|| ServiceError::ErrorResponse(StatusCode::NOT_FOUND, Some("booking not found".to_string())))?;
    Ok(Json(booking))
}

async fn booking_rooms(
    Path(id): Path<i32>,
    Extension(service): Extension<Service>,
) -> Result<impl IntoResponse, ServiceError> {
    let rooms = service.get_rooms_for_booking(id).await?;
    Ok(Json(rooms))
}

async fn booking_guests(
    Path(id): Path<i32>,
    Extension(service): Extension<Service>,
) -> Result<impl IntoResponse, ServiceError> {
    let guests = service.get_guests_for_booking(id).await?;
    Ok(Json(guests))
}

async fn add_booking_room(
    Path((id, roomid)): Path<(i32, i32)>,
    Extension(service): Extension<Service>,
) -> Result<impl IntoResponse, ServiceError> {
    service.add_booking_room(id, roomid).await?;
    Ok(StatusCode::CREATED)
}

async fn add_booking_guest(
    Path((id, personid)): Path<(i32, i32)>,
    Extension(service): Extension<Service>,
) -> Result<impl IntoResponse, ServiceError> {
    service.add_booking_guest(id, personid).await?;
    Ok(StatusCode::CREATED)
}

async fn delete_booking_room(
    Path((id, roomid)): Path<(i32, i32)>,
    Extension(service): Extension<Service>,
) -> Result<impl IntoResponse, ServiceError> {
    service.delete_booking_room(id, roomid).await?;
    Ok(StatusCode::NO_CONTENT)
}

async fn delete_booking_guest(
    Path((id, personid)): Path<(i32, i32)>,
    Extension(service): Extension<Service>,
) -> Result<impl IntoResponse, ServiceError> {
    service.delete_booking_guest(id, personid).await?;
    Ok(StatusCode::NO_CONTENT)
}

async fn add_booking(
    Json(payload): Json<BookingBody>,
    Extension(service): Extension<Service>,
) -> Result<impl IntoResponse, ServiceError> {
    let id = service.add_booking(payload.clone()).await?;
    let booking = Booking {
        id,
        arrival_date: payload.arrival_date,
        depature_date: payload.depature_date,
        cost: payload.cost,
        pension_type: payload.pension_type,
        client_id: payload.client_id,
    };
    Ok((StatusCode::CREATED, Json(booking)))
}

async fn delete_booking(
    Path(id): Path<i32>,
    Extension(service): Extension<Service>,
) -> Result<impl IntoResponse, ServiceError> {
    service.delete_booking(id).await?;
    Ok(StatusCode::NO_CONTENT)
}

async fn update_booking(
    Json(payload): Json<BookingBody>,
    Path(id): Path<i32>,
    Extension(service): Extension<Service>,
) -> Result<impl IntoResponse, ServiceError> {
    service.update_booking(id, payload).await?;
    Ok(StatusCode::NO_CONTENT)
}

pub fn routes() -> Router<BoxRoute> {
    Router::new()
        .route("/", get(bookings).post(add_booking))
        .route("/:id", get(booking).delete(delete_booking).put(update_booking))
        .route("/:id/rooms", get(booking_rooms))
        .route("/:id/rooms/:roomid", put(add_booking_room).delete(delete_booking_room))
        .route("/:id/guests", get(booking_guests))
        .route("/:id/guests/:personid", put(add_booking_guest).delete(delete_booking_guest))
        .boxed()
}